from __future__ import annotations

from enum import Enum, auto


class GitRepoPropName(Enum):
    """
    Envelope arg types (properties) used by `GitRepoLoader`.
    """

    # Short name for a repo (to avoid specifying its `git_repo_root_rel_path`):
    git_repo_alias = auto()

    # Absolute path common to multiple Git repos - what `git_repo_root_rel_path` are relative to:
    git_repo_root_abs_path = auto()

    # Git root path relative to `git_repo_root_abs_path`:
    git_repo_root_rel_path = auto()

    # Git root dir base name:
    git_repo_root_base_name = auto()

    # Categorize repo content: conf, code, ...
    git_repo_content_type = auto()

    # Git tag, for example: `v0.6.7.final`
    git_repo_tag_name = auto()

    # Git commit, for example: 52a3d54d17fe5cba6bad57abc4b6bebd881d3ce6
    git_repo_commit_id = auto()

    # Git commit, for example: 52a3d54d17fe5cba6bad57abc4b6bebd881d3ce6
    git_repo_short_commit_id = auto()

    # Git commit comment:
    git_repo_commit_message = auto()

    # Git commit author name:
    git_repo_commit_author_name = auto()

    # Git commit author email:
    git_repo_commit_author_email = auto()

    # Git commit date (UTC):
    git_repo_commit_date = auto()

    # Git commit time (UTC):
    git_repo_commit_time = auto()

    # FS_06_99_43_60: example of using non-scalar value (array|list):
    # List of categories object belongs to:
    git_repo_object_category = auto()

    def __str__(self):
        return self.name



from marshmallow import fields, RAISE

from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
from argrelay.server_spec.const_int import DEFAULT_IP_ADDRESS, DEFAULT_PORT_NUMBER

server_host_name_ = "server_host_name"
server_port_number_ = "server_port_number"


class ConnectionConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = ConnectionConfig

    server_host_name = fields.String(
        load_default = DEFAULT_IP_ADDRESS,
    )

    server_port_number = fields.Integer(
        load_default = DEFAULT_PORT_NUMBER,
    )


connection_config_desc = TypeDesc(
    dict_schema = ConnectionConfigSchema(),
    ref_name = ConnectionConfigSchema.__name__,
    dict_example = {
        server_host_name_: DEFAULT_IP_ADDRESS,
        server_port_number_: DEFAULT_PORT_NUMBER,
    },
    default_file_path = "",
)



from enum import Enum, auto


class ReservedEnvelopeClass(Enum):
    """
    Names of the `data_envelope` classes reserved by `argrelay`.
    """

    ClassUnknown = auto()

    ClassFunction = auto()

    ClassHelp = auto()

    def __str__(self):
        return self.name



from argrelay.relay_client import AbstractClientCommand
from argrelay.server_spec.CallContext import CallContext


class AbstractClientCommandFactory:

    def create_command(
        self,
        call_ctx: CallContext,
    ) -> AbstractClientCommand:
        pass



from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.relay_server.QueryEngine import QueryEngine
from argrelay.runtime_data.StaticData import StaticData


class NoopLoader(AbstractLoader):

    def update_static_data(
        self,
        static_data: StaticData,
        query_engine: QueryEngine,
    ) -> StaticData:
        return static_data



from marshmallow import fields, RAISE

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.AssignedValue import AssignedValue

arg_value_ = "arg_value"
arg_source_ = "arg_source"


class AssignedValueSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = AssignedValue

    arg_value = fields.String(
        required = True,
    )

    arg_source = fields.Enum(
        ArgSource,
        by_value = False,
        required = True,
    )


assigned_value_desc = TypeDesc(
    dict_schema = AssignedValueSchema(),
    ref_name = AssignedValueSchema.__name__,
    dict_example = {
        arg_value_: "value_1",
        arg_source_: ArgSource.ImplicitValue.name,
    },
    default_file_path = "",
)



from argrelay.handler_response.AbstractClientResponseHandler import AbstractClientResponseHandler
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.schema_response.ArgValues import ArgValues
from argrelay.schema_response.ArgValuesSchema import arg_values_desc


class ProposeArgValuesClientResponseHandler(AbstractClientResponseHandler):
    """
    This handler is supposed to be used by non-optimized implementations like:
    *   `ProposeArgValuesLocalClientCommand`
    *   `ProposeArgValuesRemoteClientCommand`

    Optimized `ProposeArgValuesRemoteOptimizedClientCommand` prints suggestion without using marshmallow `Schema`-s.
    """

    def __init__(
        self,
    ):
        super().__init__(
        )

    def handle_response(
        self,
        response_dict: dict,
    ):
        arg_values: ArgValues = arg_values_desc.dict_schema.load(response_dict)
        ElapsedTime.measure("after_object_creation")
        self.render_values(arg_values)

    @staticmethod
    def render_values(
        arg_values: ArgValues,
    ):
        print("\n".join(arg_values.arg_values))



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Union

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.AssignedValue import AssignedValue


@dataclass
class EnvelopeContainer:
    """
    Combination of relevant config, user input, and the last `data_envelope` found based on that.

    See also :class:`EnvelopeContainerSchema`.
    """

    search_control: SearchControl = field(default_factory = lambda: SearchControl())
    """
    A specs how to query `data_envelope` for this `EnvelopeContainer`.
    """

    data_envelopes: list[dict] = field(default_factory = lambda: [])
    """
    These `data_envelope`-s are based on `search_control` for this `EnvelopeContainer` (and command line input).

    It contains only 0 or 1 `data_envelope` for performance reasons in case of:
    *   `ServerAction.ProposeArgValues` (Tab-completion)
    *   `ServerAction.DescribeLineArgs` (Alt+Shift+Q)
    See `QueryEngine.query_prop_values` and `QueryResult.data_envelopes`.

    It contains all N `data_envelope`-s in case of `ServerAction.RelayLineArgs`.
    See `QueryResult.data_envelopes`.
    """

    found_count: int = field(default = 0)
    """
    Counter of `data_envelope`-s found in the last query.
    """

    used_arg_bucket: Union[int, None] = field(default = None)
    """
    FS_97_64_39_94 `arg_bucket` (index) where args were consumed from for the current `envelope_container`.
    """

    # TODO: Maybe rename to `context_types_to_values`?
    # TODO: Part of (or actually is?) `args_context`: FS_62_25_92_06:
    assigned_types_to_values: dict[str, AssignedValue] = field(default_factory = lambda: {})
    """
    All assigned args (from interpreted tokens) mapped as type:value which belong to `data_envelope`.
    """

    # TODO: Part of (or not?) `args_context` (FS_62_25_92_06) to support FS_13_51_07_97 (single out implicit values):
    remaining_types_to_values: dict[str, list[str]] = field(default_factory = lambda: {})
    """
    All arg values per arg type left as options to match arg value given on the command line.

    When arg value from command line matches one of the values, this arg type moves to `assigned_types_to_values`.
    """

    filled_types_to_values_hidden_by_defaults: dict[str, list[str]] = field(default_factory = lambda: {})
    """
    FS_72_53_55_13: values hidden by defaults:
    This dict stores values which were hidden by applying defaults (see FS_72_40_53_00 fill control).
    Only keys for those `arg_type`-s which have `ArgSource.InitValue` in `assigned_types_to_values` are listed here.
    """

    def populate_implicit_arg_values(
        self,
    ) -> bool:
        """
        When possible arg_values are singled out, assign them as `ArgSource.ImplicitValue`.

        When `data_envelope` is singled out, all remaining single-value `arg_type`-s become `ArgSource.ImplicitValue`.

        Implements FS_13_51_07_97 singled out implicit values.

        Return:
        *   True if any value was assigned.
        *   False otherwise.
        """

        any_assigned = False
        # Filter as in: FS_31_70_49_15 search control:
        for arg_type in self.search_control.keys_to_types_dict.values():
            if arg_type not in self.assigned_types_to_values:
                if arg_type in self.remaining_types_to_values:
                    arg_values = self.remaining_types_to_values[arg_type]

                    # FS_06_99_43_60 list arg value:
                    # When a `data_envelope` is singled out, its array arg value will not be set as
                    # singled out via `ArgSource.ImplicitValue` (user can still select them explicitly).
                    # This is deliberate for now as the selection of specific value out of the array can be used
                    # by delegators to implement difference in behavior.

                    if len(arg_values) == 1:
                        del self.remaining_types_to_values[arg_type]
                        self.assigned_types_to_values[arg_type] = AssignedValue(
                            # single value:
                            arg_values[0],
                            ArgSource.ImplicitValue,
                        )
                        any_assigned = True
        return any_assigned



from __future__ import annotations

import subprocess

from argrelay.custom_integ.GitRepoEnvelopeClass import GitRepoEnvelopeClass
from argrelay.custom_integ.GitRepoPropName import GitRepoPropName
from argrelay.custom_integ.value_constants import (
    goto_git_repo_func_,
    desc_git_commit_func_,
    desc_git_tag_func_,
)
from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.misc_helper_common import eprint
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator, get_func_id_from_invocation_input
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext, function_container_ipos_
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_payload_,
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_config_interp.SearchControlSchema import populate_search_control
from argrelay.schema_response.InvocationInput import InvocationInput

repo_container_ipos_ = 1


class GitRepoDelegator(AbstractDelegator):
    """
    Implements FS_67_16_61_97 git_plugin.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        repo_search_control = populate_search_control(
            class_to_collection_map,
            GitRepoEnvelopeClass.ClassGitRepo.name,
            [
                {"category": GitRepoPropName.git_repo_object_category.name},
                {"alias": GitRepoPropName.git_repo_alias.name},
                {"content": GitRepoPropName.git_repo_content_type.name},
                {"name": GitRepoPropName.git_repo_root_base_name.name},
                {"path": GitRepoPropName.git_repo_root_rel_path.name},
                {"base": GitRepoPropName.git_repo_root_abs_path.name},
            ],
        )

        tag_search_control = populate_search_control(
            class_to_collection_map,
            GitRepoEnvelopeClass.ClassGitTag.name,
            [
                {"category": GitRepoPropName.git_repo_object_category.name},

                {"alias": GitRepoPropName.git_repo_alias.name},
                {"content": GitRepoPropName.git_repo_content_type.name},
                {"name": GitRepoPropName.git_repo_root_base_name.name},
                {"path": GitRepoPropName.git_repo_root_rel_path.name},
                {"base": GitRepoPropName.git_repo_root_abs_path.name},

                {"date": GitRepoPropName.git_repo_commit_date.name},
                {"tag": GitRepoPropName.git_repo_tag_name.name},
                {"time": GitRepoPropName.git_repo_commit_time.name},
                {"hex": GitRepoPropName.git_repo_short_commit_id.name},
            ],
        )

        commit_search_control = populate_search_control(
            class_to_collection_map,
            GitRepoEnvelopeClass.ClassGitCommit.name,
            [
                {"category": GitRepoPropName.git_repo_object_category.name},

                {"alias": GitRepoPropName.git_repo_alias.name},
                {"content": GitRepoPropName.git_repo_content_type.name},
                {"name": GitRepoPropName.git_repo_root_base_name.name},
                {"path": GitRepoPropName.git_repo_root_rel_path.name},
                {"base": GitRepoPropName.git_repo_root_abs_path.name},

                {"date": GitRepoPropName.git_repo_commit_date.name},
                {"email": GitRepoPropName.git_repo_commit_author_email.name},
                {"time": GitRepoPropName.git_repo_commit_time.name},
                {"hex": GitRepoPropName.git_repo_short_commit_id.name},
            ],
        )

        func_envelopes = []

        given_function_envelope = {
            instance_data_: {
                func_id_: goto_git_repo_func_,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    repo_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedPropName.help_hint.name: "Goto Git repository (`cd` to its path)",
            ReservedPropName.func_state.name: FuncState.fs_beta.name,
            ReservedPropName.func_id.name: goto_git_repo_func_,
        }
        func_envelopes.append(given_function_envelope)

        given_function_envelope = {
            instance_data_: {
                func_id_: desc_git_tag_func_,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    tag_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedPropName.help_hint.name: "Describe Git tag",
            ReservedPropName.func_state.name: FuncState.fs_demo.name,
            ReservedPropName.func_id.name: desc_git_tag_func_,
        }
        func_envelopes.append(given_function_envelope)

        given_function_envelope = {
            instance_data_: {
                func_id_: desc_git_commit_func_,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    commit_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedPropName.help_hint.name: "Describe Git commit",
            ReservedPropName.func_state.name: FuncState.fs_demo.name,
            ReservedPropName.func_id.name: desc_git_commit_func_,
        }
        func_envelopes.append(given_function_envelope)

        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:

        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        # The first envelope (`DataEnvelopeSchema`) is assumed to be of
        # `ReservedEnvelopeClass.ClassFunction` with `FunctionEnvelopeInstanceDataSchema` for its `instance_data`:
        function_container = interp_ctx.envelope_containers[function_container_ipos_]
        delegator_plugin_instance_id = (
            function_container.data_envelopes[0][instance_data_]
            [delegator_plugin_instance_id_]
        )
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.plugin_instance_entries[
                delegator_plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        if get_func_id_from_invocation_input(invocation_input) == goto_git_repo_func_:
            repo_envelope = invocation_input.envelope_containers[repo_container_ipos_].data_envelopes[0]
            repo_root_abs_path = repo_envelope[envelope_payload_][repo_root_abs_path_]
            eprint(f"INFO: starting subshell in: {repo_root_abs_path}")
            # List Git repo dir:
            sub_proc = subprocess.run(
                [
                    "bash",
                    "-l",
                ],
                cwd = repo_root_abs_path
            )
            exit_code = sub_proc.returncode
            if exit_code != 0:
                raise RuntimeError
        if get_func_id_from_invocation_input(invocation_input) == desc_git_tag_func_:
            raise RuntimeError("not implemented")
        if get_func_id_from_invocation_input(invocation_input) == desc_git_commit_func_:
            raise RuntimeError("not implemented")


repo_root_abs_path_: str = "repo_root_abs_path"



from __future__ import annotations

from copy import deepcopy

from argrelay.composite_tree.CompositeInfoType import CompositeInfoType
from argrelay.composite_tree.CompositeTreeWalker import extract_jump_tree, extract_func_tree
from argrelay.composite_tree.DictTreeWalker import DictTreeWalker, normalize_tree, sequence_starts_with
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.misc_helper_common import eprint
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay.plugin_interp.FuncTreeInterp import FuncTreeInterp, func_init_control_, func_search_control_
from argrelay.plugin_interp.FuncTreeInterpFactoryConfigSchema import (
    func_selector_tree_,
    func_tree_interp_config_desc,
    jump_tree_,
)
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import init_envelop_collections
from argrelay.schema_config_interp.InitControlSchema import init_types_to_values_
from argrelay.schema_config_interp.SearchControlSchema import (
    keys_to_types_list_,
    populate_search_control,
)
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_enabled_, plugin_dependencies_

tree_path_selector_prefix_ = "tree_path_selector_"
tree_path_selector_key_prefix_ = "s"


class FuncTreeInterpFactory(AbstractInterpFactory):
    """
    Implements FS_26_43_73_72 func tree.
    Implements FS_91_88_07_23 jump tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

        self._compare_config_with_composite_tree()

        # FS_26_43_73_72 func tree: func id to list of its absolute paths populated by `load_func_envelopes`.
        # These func tree paths are absolute within FS_33_76_82_84 composite tree.
        # Each func id can be attached to more than one leaf (hence, there is a list of paths to that func id).
        self.func_ids_to_func_abs_paths: dict[str, list[list[str]]] = {}

        # FS_91_88_07_23 jump tree
        self.plugin_config_dict.setdefault(jump_tree_, {})
        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.jump_tree,
            self.plugin_config_dict[jump_tree_],
        )
        self.paths_to_jump: dict[tuple[str, ...], tuple[str, ...]] = dict_tree_walker.build_paths_to_paths()
        """
        Implements FS_91_88_07_23 jump tree:
        for given `interp_tree_abs_path` (FS_01_89_09_24) selects next `interp_tree_abs_path`.
        """

    # TODO_10_72_28_05: This will go away together with switch to FS_33_76_82_84 composite tree config:
    def _compare_config_with_composite_tree(
        self,
    ):
        expected_dict = self.plugin_config_dict[jump_tree_]
        actual_dict = extract_jump_tree(
            self.server_config.server_plugin_control.composite_forest,
        )
        eprint(f"expected_dict: {expected_dict}")
        eprint(f"actual_dict: {actual_dict}")
        assert expected_dict == actual_dict

    # TODO_10_72_28_05: This will go away together with switch to FS_33_76_82_84 composite tree config:
    def _compare_config_with_composite_tree_func_tree(
        self,
        expected_func_selector_tree_dict: dict,
    ):
        expected_dict = normalize_tree(expected_func_selector_tree_dict)
        actual_dict = normalize_tree(extract_func_tree(
            self.server_config.server_plugin_control.composite_forest,
            self.plugin_instance_id,
        ))
        eprint(f"expected_dict: {expected_dict}")
        eprint(f"actual_dict: {actual_dict}")
        assert expected_dict == actual_dict

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return func_tree_interp_config_desc.dict_from_input_dict(plugin_config_dict)

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        self.interp_tree_abs_paths.extend(
            this_plugin_instance_interp_tree_abs_paths,
        )

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> list[str]:
        """
        To implement FS_26_43_73_72 func tree, this plugin loads func `data_envelope`-s automatically.

        It loops through func `data_envelope`-s and populates their tree path search props
        according to configured `func_selector_tree` where each function is plugged into.
        """

        mapped_func_ids: list[str] = super().load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )
        interp_tree_node_config_dict = self.interp_tree_abs_paths_to_node_configs[interp_tree_abs_path]

        self._compare_config_with_composite_tree_func_tree(
            interp_tree_node_config_dict[func_selector_tree_],
        )

        dict_tree_walker = DictTreeWalker(
            CompositeInfoType.func_tree,
            interp_tree_node_config_dict[func_selector_tree_],
        )
        self.func_ids_to_func_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()

        # `func_envelope` (2-level map) per `func_id` per `func_abs_path`
        func_id_to_func_abs_path_to_func_envelope: dict[str, dict[tuple[str, ...], dict]] = {}
        # Loop through func `data_envelope`-s from all delegators:
        for func_id, func_envelope in func_ids_to_func_envelopes.items():
            if func_id not in self.func_ids_to_func_abs_paths:
                # TODO: TODO_19_67_22_89: remove `ignored_func_ids_list` - load as `FuncState.fs_unplugged`:
                # This `func_id` is not plugged into given tree.
                # It will be loaded as `FuncState.fs_unplugged`.
                continue
            func_abs_paths = self.func_ids_to_func_abs_paths[func_id]
            for func_abs_path in func_abs_paths:
                assert (
                    func_id not in func_id_to_func_abs_path_to_func_envelope
                    or
                    tuple(func_abs_path) not in func_id_to_func_abs_path_to_func_envelope[func_id]
                )
                if self._is_best_matching_abs_tree_path(func_abs_path, interp_tree_abs_path):
                    # This `func_id` is mapped into the `func_tree` under this plugin:
                    func_id_to_func_abs_path_to_func_envelope.setdefault(
                        func_id,
                        {},
                    )[tuple(func_abs_path)] = deepcopy(func_envelope)
                    mapped_func_ids.append(func_id)

        # Ensure every `func_id` mapped into the tree is part of `func_envelope`-s loaded from delegators:
        for func_id in self.func_ids_to_func_abs_paths.keys():
            if func_id not in mapped_func_ids:
                raise RuntimeError(
                    f"plugin_instance_id=`{self.plugin_instance_id}`: func_id=`{func_id}` is not published by any enabled delegator activated so far - please check `{plugin_enabled_}` and `{plugin_dependencies_}`"
                )

        self.populate_func_envelope_collection(
            func_id_to_func_abs_path_to_func_envelope,
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )

        return mapped_func_ids

    def populate_func_envelope_collection(
        self,
        func_id_to_func_abs_path_to_func_envelope: dict[str, dict[tuple[str, ...], dict]],
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
    ):
        self.populate_init_control(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )
        self.populate_search_control(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
        )
        (
            interp_tree_abs_path_func_envelopes,
            prop_names,
        ) = self.populate_func_tree_props(
            interp_tree_abs_path,
            interp_tree_node_config_dict,
            func_id_to_func_abs_path_to_func_envelope,
        )
        class_to_collection_map: dict = self.server_config.class_to_collection_map
        class_names = [
            ReservedEnvelopeClass.ClassFunction.name,
        ]
        init_envelop_collections(
            self.server_config,
            class_names,
            lambda collection_name, class_name: (
                [
                    ReservedPropName.envelope_class.name,
                ] + prop_names
            )
        )
        envelope_collection = self.server_config.static_data.envelope_collections[
            class_to_collection_map[ReservedEnvelopeClass.ClassFunction.name]
        ]
        # Write func envelopes into `StaticData` (as if it is a loader plugin):
        envelope_collection.data_envelopes.extend(interp_tree_abs_path_func_envelopes)

    def _is_best_matching_abs_tree_path(
        self,
        func_abs_path,
        given_interp_tree_abs_path,
    ) -> bool:
        if sequence_starts_with(func_abs_path, given_interp_tree_abs_path):
            best_matching_abs_tree_path = given_interp_tree_abs_path
            max_matched_len = len(best_matching_abs_tree_path)
            for interp_tree_abs_path in self.interp_tree_abs_paths:
                if given_interp_tree_abs_path == interp_tree_abs_path:
                    continue
                if sequence_starts_with(func_abs_path, interp_tree_abs_path):
                    if max_matched_len < len(interp_tree_abs_path):
                        best_matching_abs_tree_path = interp_tree_abs_path
                        max_matched_len = len(best_matching_abs_tree_path)
                        break
                    elif max_matched_len == len(interp_tree_abs_path):
                        # There should not be two plugin paths leading to the same func:
                        raise RuntimeError()
                    else:
                        continue
            return given_interp_tree_abs_path == best_matching_abs_tree_path
        else:
            return False

    # noinspection PyMethodMayBeStatic
    def populate_init_control(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict: dict,
    ):
        interp_tree_node_config_dict[func_init_control_] = {
            init_types_to_values_: {},
        }

        init_types_to_values = interp_tree_node_config_dict[func_init_control_][init_types_to_values_]
        for sel_ipos, path_step in enumerate(interp_tree_abs_path):
            init_types_to_values[f"{func_envelope_path_step_prop_name(sel_ipos)}"] = path_step

    def populate_search_control(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
    ):
        """
        Provide `func_search_control` based on `func_ids_to_func_abs_paths`.
        """

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        interp_tree_node_config_dict[func_search_control_] = populate_search_control(
            class_to_collection_map,
            ReservedEnvelopeClass.ClassFunction.name,
            [],
        )

        # `func_search_control` should include keys from the interp tree abs path:
        keys_to_types_list = interp_tree_node_config_dict[func_search_control_][keys_to_types_list_]

        # Include func tree path:
        max_len = max_path_len(self.func_ids_to_func_abs_paths)
        for sel_ipos in range(max_len):
            keys_to_types_list.append({
                f"{path_step_key_arg_name(sel_ipos)}": f"{func_envelope_path_step_prop_name(sel_ipos)}"
            })

        # Include other fields:
        keys_to_types_list.append({
            "state": ReservedPropName.func_state.name
        })
        keys_to_types_list.append({
            "id": ReservedPropName.func_id.name
        })

    def populate_func_tree_props(
        self,
        interp_tree_abs_path: tuple[str, ...],
        interp_tree_node_config_dict,
        func_envelopes_index: dict[str, dict[tuple[str, ...], dict]],
    ) -> tuple[list[dict], list[str]]:
        """
        Populates func tree properties for each func `data_envelope` based on its path in func tree.

        For each func envelope:
        *   Select its place on the func tree (provided by plugin config).
        *   Populate the path as envelope props with `tree_path_selector_prefix_`.
        """
        interp_tree_abs_path_func_envelopes: list[dict] = []
        prop_names: list[str] = []
        for func_id in self.func_ids_to_func_abs_paths:
            func_abs_paths: list[list[str]] = self.func_ids_to_func_abs_paths[func_id]

            for func_abs_path in func_abs_paths:

                if (
                    func_id not in func_envelopes_index
                    or
                    tuple(func_abs_path) not in func_envelopes_index[func_id]
                ):
                    continue

                func_envelope = func_envelopes_index[func_id][tuple(func_abs_path)]

                # Include func tree path:
                for sel_ipos, path_step_id in enumerate(func_abs_path):
                    prop_name = func_envelope_path_step_prop_name(sel_ipos)
                    func_envelope[prop_name] = path_step_id
                    prop_names.append(prop_name)

                interp_tree_abs_path_func_envelopes.append(func_envelope)

        return (
            interp_tree_abs_path_func_envelopes,
            prop_names,
        )

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> AbstractInterp:
        return FuncTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[interp_ctx.interp_tree_abs_path],
            interp_ctx,
            self.func_ids_to_func_abs_paths,
            self.paths_to_jump,
        )


def func_envelope_path_step_prop_name(
    path_step_ipos: int,
) -> str:
    return f"{tree_path_selector_prefix_}{path_step_ipos}"


def path_step_key_arg_name(
    path_step_ipos: int,
) -> str:
    return f"{tree_path_selector_key_prefix_}{path_step_ipos}"


def max_path_len(
    func_ids_to_func_abs_paths: dict[str, list[list[str]]],
):
    """
    Find maximum path length.
    """
    max_len: int = 0
    for func_rel_paths in func_ids_to_func_abs_paths.values():
        for func_rel_path in func_rel_paths:
            max_len = max(max_len, len(func_rel_path))
    return max_len



from argrelay.client_command_local.AbstractLocalClientCommand import AbstractLocalClientCommand
from argrelay.enum_desc.ServerAction import ServerAction
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.relay_client.AbstractClientCommandFactory import AbstractClientCommandFactory
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_data.PluginConfig import PluginConfig
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_server.ServerConfigSchema import server_config_desc
from argrelay.schema_config_plugin.PluginConfigSchema import plugin_config_desc
from argrelay.server_spec.CallContext import CallContext


class LocalClientCommandFactory(AbstractClientCommandFactory):
    local_server: LocalServer = None
    """
    This instance of `LocalServer` allows reusing server for multiple `LocalClient` invocation in tests.
    For example, to avoid re-loading TD_38_03_48_51 large data sets.
    """

    def __init__(self):
        self._start_local_server()

    def _start_local_server(self):
        if not LocalClientCommandFactory.local_server:
            self.server_config: ServerConfig = server_config_desc.obj_from_default_file()
            self.plugin_config: PluginConfig = plugin_config_desc.obj_from_default_file()
            ElapsedTime.measure("after_server_config_load")
            LocalClientCommandFactory.local_server = LocalServer(
                self.server_config,
                self.plugin_config,
            )
            LocalClientCommandFactory.local_server.start_local_server()
            ElapsedTime.measure("after_local_server_start")
        else:
            self.server_config: ServerConfig = LocalClientCommandFactory.local_server.server_config

    def create_command(
        self,
        call_ctx: CallContext,
    ) -> AbstractLocalClientCommand:
        if call_ctx.server_action is ServerAction.ProposeArgValues:
            from argrelay.client_command_local.ProposeArgValuesLocalClientCommand import (
                ProposeArgValuesLocalClientCommand
            )
            return ProposeArgValuesLocalClientCommand(
                call_ctx,
                self.server_config,
                self.local_server,
            )
        if call_ctx.server_action is ServerAction.DescribeLineArgs:
            from argrelay.client_command_local.DescribeLineArgsLocalClientCommand import (
                DescribeLineArgsLocalClientCommand
            )
            return DescribeLineArgsLocalClientCommand(
                call_ctx,
                self.server_config,
                self.local_server,
            )
        if call_ctx.server_action is ServerAction.RelayLineArgs:
            from argrelay.client_command_local.RelayLineArgsLocalClientCommand import (
                RelayLineArgsLocalClientCommand
            )
            return RelayLineArgsLocalClientCommand(
                call_ctx,
                self.server_config,
                self.local_server,
            )
        raise RuntimeError



from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper_common.TypeDesc import TypeDesc

func_selector_tree_ = "func_selector_tree"

jump_tree_ = "jump_tree"


# TODO_40_10_18_32: add custom base to all schemas:
class FuncTreeInterpFactoryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    # TODO_79_67_28_83: Express recursive dict schema:
    # This is a tree (`dict`) of arbitrary depth with `str` leaves.
    # Ideally, this should be defined as nested `dict`,
    # but it is unknown how to do it in marshmallow.
    # Implements FS_26_43_73_72 func tree.
    func_selector_tree = fields.Raw(
        required = True,
    )

    # TODO_79_67_28_83: Express recursive dict schema:
    # This is a tree (`dict`) of arbitrary depth with `list[str]` leaves.
    # Ideally, this should be defined as nested `dict`,
    # but it is unknown how to do it in marshmallow.
    # Implements FS_91_88_07_23 jump tree.
    jump_tree = fields.Raw(
        required = False,
    )


func_tree_interp_config_example = {
    func_selector_tree_: {
        "list": {
            "repo": "list_repo_func",
            "service": "list_service_func",
        },
        "diff": {
            "service": "diff_service_func",
        },
        "goto": {
            "repo": "goto_git_repo_func",
            "service": "goto_service_func",
        },
    },
    jump_tree_: [
        "relay_demo",
    ],
}

func_tree_interp_config_desc = TypeDesc(
    dict_schema = FuncTreeInterpFactoryConfigSchema(),
    ref_name = FuncTreeInterpFactoryConfigSchema.__name__,
    dict_example = func_tree_interp_config_example,
    default_file_path = "",
)



from __future__ import annotations

from enum import IntEnum

from argrelay.enum_desc.CompType import CompType


class CompScope(IntEnum):
    """
    Provides scope of completion.

    Roughly this means:
    *   size of the proposed list of values
    *   allowed sources of values for completion
    """

    ScopeUnknown = 0

    ScopeInitial = 1

    ScopeSubsequent = 2

    @classmethod
    def from_comp_type(
        cls,
        comp_type: CompType,
    ) -> CompScope:
        """
        Original idea is to decouple client and server.

        Completion type based on action in Bash (see `CompType`) seen by client includes artificial cases
        like `CompType.DescribeArgs` and `CompType.InvokeAction` which are already orthogonally modeled via
        `ServerAction` and should not be seen in `CompScope`.

        Completion type based on initial and subsequent request fall into only two categories (so far)
        and do not need to be distinguished further on the server side:
        *   initial request: `CompType.PrefixShown` == `CompType.PrefixHidden` == `CompType.MenuCompletion`
        *   subsequent request: `CompType.SubsequentHelp`
        """
        if comp_type in [
            CompType.PrefixShown,
            CompType.PrefixHidden,
            CompType.MenuCompletion,
        ]:
            return CompScope.ScopeInitial

        if comp_type is CompType.SubsequentHelp:
            return CompScope.ScopeSubsequent

        return CompScope.ScopeUnknown



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.composite_tree.CompositeNode import ZeroArgNode


@dataclass
class CompositeForest:
    """
    FS_33_76_82_84 composite tree combined into forest.

    See also :class:`CompositeForestSchema`.
    """

    tree_roots: dict[str, ZeroArgNode] = field()



from __future__ import annotations

from argrelay.custom_integ.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay.custom_integ.ServiceLoaderConfigSchema import (
    service_loader_config_desc,
    test_data_ids_to_load_,
)
from argrelay.custom_integ.ServicePropName import ServicePropName
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.misc_helper_common import eprint
from argrelay.plugin_loader.AbstractLoader import AbstractLoader
from argrelay.relay_server.QueryEngine import QueryEngine
from argrelay.runtime_data.EnvelopeCollection import EnvelopeCollection
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import init_envelop_collections
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_payload_,
    envelope_id_,
)
from argrelay.test_infra import test_data_


# noinspection PyPep8Naming
class ServiceLoader(AbstractLoader):
    object_multiplier: int = 10
    """
    Used by TD_38_03_48_51 to generate large data set.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

    def load_config(
        self,
        plugin_config_dict: dict,
    ) -> dict:
        return service_loader_config_desc.dict_from_input_dict(plugin_config_dict)

    def update_static_data(
        self,
        static_data: StaticData,
        query_engine: QueryEngine,
    ) -> StaticData:

        static_data = self.load_data_envelopes(static_data)

        return static_data

    # noinspection PyMethodMayBeStatic
    def load_data_envelopes(
        self,
        static_data: StaticData,
    ) -> StaticData:
        """
        The loader writes samples into `static_data` simply from code (without any data source).
        """

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        class_names = [
            ServiceEnvelopeClass.ClassCluster.name,
            ServiceEnvelopeClass.ClassHost.name,
            ServiceEnvelopeClass.ClassService.name,
            ServiceEnvelopeClass.ClassAccessType.name,
        ]

        init_envelop_collections(
            self.server_config,
            class_names,
            # Same index fields for all collections (can be fine-tuned later):
            lambda collection_name, class_name: [enum_item.name for enum_item in ServicePropName]
        )

        # Select `data_envelope` lists used by each collection name
        cluster_envelopes = static_data.envelope_collections[
            class_to_collection_map[ServiceEnvelopeClass.ClassCluster.name]
        ].data_envelopes
        host_envelopes = static_data.envelope_collections[
            class_to_collection_map[ServiceEnvelopeClass.ClassHost.name]
        ].data_envelopes
        service_envelopes = static_data.envelope_collections[
            class_to_collection_map[ServiceEnvelopeClass.ClassService.name]
        ].data_envelopes
        access_envelopes = static_data.envelope_collections[
            class_to_collection_map[ServiceEnvelopeClass.ClassAccessType.name]
        ].data_envelopes

        self.populate_common_access_type(access_envelopes)
        self.populate_TD_63_37_05_36_default(
            cluster_envelopes,
            host_envelopes,
            service_envelopes,
        )
        self.populate_TD_76_09_29_31_overlapped(
            cluster_envelopes,
            host_envelopes,
            service_envelopes,
        )
        self.populate_TD_99_99_88_75_mutually_exclusive(
            cluster_envelopes,
            host_envelopes,
            service_envelopes,
        )
        self.populate_TD_38_03_48_51_large_generated(
            cluster_envelopes,
            host_envelopes,
            service_envelopes,
        )
        self.populate_TD_43_24_76_58_single(
            cluster_envelopes,
            host_envelopes,
            service_envelopes,
        )
        self.populate_TD_39_25_11_76_missing_props(
            cluster_envelopes,
            host_envelopes,
            service_envelopes,
        )

        self.generate_envelope_id(cluster_envelopes + host_envelopes + service_envelopes)

        self.generate_help_hints(
            class_to_collection_map,
            static_data,
        )

        return static_data

    # noinspection PyMethodMayBeStatic
    def generate_envelope_id(
        self,
        data_envelopes: list[dict],
    ):
        for data_envelope in data_envelopes:
            if envelope_id_ not in data_envelope:
                if data_envelope[ReservedPropName.envelope_class.name] == ServiceEnvelopeClass.ClassHost.name:
                    data_envelope[envelope_id_] = (
                        data_envelope[ServicePropName.cluster_name.name]
                        + "." +
                        data_envelope[ServicePropName.host_name.name]
                    )
                if data_envelope[ReservedPropName.envelope_class.name] == ServiceEnvelopeClass.ClassService.name:
                    data_envelope[envelope_id_] = (
                        data_envelope[ServicePropName.cluster_name.name]
                        + "." +
                        data_envelope[ServicePropName.host_name.name]
                        + "." +
                        data_envelope[ServicePropName.service_name.name]
                        + "." +
                        data_envelope[ServicePropName.run_mode.name]
                    )

    @staticmethod
    def generate_help_hints(
        class_to_collection_map: dict,
        static_data: StaticData,
    ):
        """
        This demos FS_71_87_33_52 help_hint for `ServicePropName.ip_address`.

        It simply generates `data_envelope`-s of `ReservedEnvelopeClass.ClassHelp` for
        values of `ServicePropName.ip_address` equal to corresponding `ServicePropName.host_name`.
        """

        class_to_collection_map.setdefault(
            ReservedEnvelopeClass.ClassHelp.name,
            ReservedEnvelopeClass.ClassHelp.name,
        )
        help_hint_envelope_collection = static_data.envelope_collections.setdefault(
            class_to_collection_map[ReservedEnvelopeClass.ClassHelp.name],
            EnvelopeCollection(
                index_props = [],
                data_envelopes = [],
            ),
        )
        help_hint_index_props = help_hint_envelope_collection.index_props
        help_hint_envelopes = help_hint_envelope_collection.data_envelopes

        # Init index fields (if they do not exist):
        for help_hint_index_prop in [
            ReservedPropName.envelope_class.name,
            ReservedPropName.arg_type.name,
            ReservedPropName.arg_value.name,
            # `ReservedPropName.help_hint` is not indexed (and uses as search param) - instead, it is a search result:
            # ReservedPropName.help_hint.name,
        ]:
            if help_hint_index_prop not in help_hint_index_props:
                help_hint_index_props.append(help_hint_index_prop)

        # Generating
        host_envelopes = static_data.envelope_collections[
            class_to_collection_map[ServiceEnvelopeClass.ClassHost.name]
        ].data_envelopes

        for host_envelope in host_envelopes:
            # This `if`-filter is not necessary until non-host-class-envelopes
            # get stored into the same collection:
            if host_envelope[ReservedPropName.envelope_class.name] == ServiceEnvelopeClass.ClassHost.name:
                if ServicePropName.ip_address.name in host_envelope:
                    help_hint_envelopes.append({
                        f"{ReservedPropName.envelope_class.name}": f"{ReservedEnvelopeClass.ClassHelp.name}",
                        f"{ReservedPropName.arg_type.name}": f"{ServicePropName.ip_address.name}",
                        f"{ReservedPropName.arg_value.name}": f"{host_envelope[ServicePropName.ip_address.name]}",
                        f"{ReservedPropName.help_hint.name}": f"{host_envelope[ServicePropName.host_name.name]}",
                    })

    def is_test_data_allowed(
        self,
        test_data_id: str,
    ) -> bool:
        if test_data_id in self.plugin_config_dict[test_data_ids_to_load_]:
            return True
        return False

    @classmethod
    def populate_common_access_type(
        cls,
        data_envelopes: list,
    ):

        data_envelopes.extend([

            ############################################################################################################
            # `access_type`: FS_24_50_40_64

            {
                envelope_payload_: {
                },
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassAccessType.name,
                ServicePropName.access_type.name: "ro",
            },
            {
                envelope_payload_: {
                },
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassAccessType.name,
                ServicePropName.access_type.name: "rw",
            },
        ])

    def populate_TD_63_37_05_36_default(
        self,
        cluster_envelopes: list[dict],
        host_envelopes: list[dict],
        service_envelopes: list[dict],
    ):
        """
        Populates TD_63_37_05_36 # demo
        """
        if not self.is_test_data_allowed("TD_63_37_05_36"):
            return

        cluster_envelopes.extend([

            ############################################################################################################
            # TD_63_37_05_36 # demo: clusters

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-apac-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-emea-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-amer-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-apac-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-emea-downstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-amer-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-amer-downstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "prod-apac-downstream",
            },
        ])

        host_envelopes.extend([
            ############################################################################################################
            # TD_63_37_05_36 # demo: hosts

            # dev

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-apac-upstream",
                ServicePropName.data_center.name: "dc.01",
                ServicePropName.host_name.name: "zxcv-du",
                ServicePropName.ip_address.name: "ip.192.168.1.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.data_center.name: "dc.11",
                ServicePropName.host_name.name: "zxcv-dd",
                ServicePropName.ip_address.name: "ip.172.16.1.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.data_center.name: "dc.01",
                ServicePropName.host_name.name: "poiu-dd",
                ServicePropName.ip_address.name: "ip.192.168.1.3",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-emea-upstream",
                ServicePropName.data_center.name: "dc.22",
                ServicePropName.host_name.name: "asdf-du",
                ServicePropName.ip_address.name: "ip.172.16.2.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.data_center.name: "dc.02",
                ServicePropName.host_name.name: "xcvb-dd",
                ServicePropName.ip_address.name: "ip.192.168.2.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-amer-upstream",
                ServicePropName.data_center.name: "dc.03",
                ServicePropName.host_name.name: "qwer-du",
                ServicePropName.ip_address.name: "ip.192.168.3.1",
            },

            # qa

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-apac-upstream",
                ServicePropName.data_center.name: "dc.04",
                ServicePropName.host_name.name: "hjkl-qu",
                ServicePropName.ip_address.name: "ip.192.168.4.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-apac-upstream",
                ServicePropName.data_center.name: "dc.44",
                ServicePropName.host_name.name: "poiu-qu",
                ServicePropName.ip_address.name: "ip.172.16.4.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-amer-upstream",
                ServicePropName.data_center.name: "dc.06",
                ServicePropName.host_name.name: "rtyu-qu",
                ServicePropName.ip_address.name: "ip.192.168.6.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-amer-upstream",
                ServicePropName.data_center.name: "dc.06",
                ServicePropName.host_name.name: "rt-qu",
                ServicePropName.ip_address.name: "ip.192.168.6.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-amer-downstream",
                ServicePropName.data_center.name: "dc.06",
                ServicePropName.host_name.name: "sdfgh-qd",
                ServicePropName.ip_address.name: "ip.192.168.6.3",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-amer-downstream",
                ServicePropName.data_center.name: "dc.06",
                ServicePropName.host_name.name: "sdfgb-qd",
                ServicePropName.ip_address.name: "ip.192.168.6.4",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-amer-downstream",
                ServicePropName.data_center.name: "dc.66",
                ServicePropName.host_name.name: "sdfg-qd",
                ServicePropName.ip_address.name: "ip.172.16.6.5",
            },

            # prod

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "qwer-pd-1",
                ServicePropName.ip_address.name: "ip.192.168.7.1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "qwer-pd-3",
                ServicePropName.ip_address.name: "ip.192.168.7.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.77",
                ServicePropName.host_name.name: "qwer-pd-2",
                ServicePropName.ip_address.name: "ip.172.16.7.2",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "prod-apac-downstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "wert-pd-1",
                ServicePropName.ip_address.name: "ip.192.168.7.3",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "prod-apac-downstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "wert-pd-2",
                ServicePropName.ip_address.name: "ip.192.168.7.4",
            },
        ])

        service_envelopes.extend([
            ############################################################################################################
            # TD_63_37_05_36 # demo: services

            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-apac-upstream",
                ServicePropName.data_center.name: "dc.01",
                ServicePropName.host_name.name: "zxcv-du",
                ServicePropName.ip_address.name: "ip.192.168.1.1",
                ServicePropName.service_name.name: "s_a",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "aaa",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-apac-upstream",
                ServicePropName.data_center.name: "dc.01",
                ServicePropName.host_name.name: "zxcv-du",
                ServicePropName.ip_address.name: "ip.192.168.1.1",
                ServicePropName.service_name.name: "s_b",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "bbb",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-apac-upstream",
                ServicePropName.data_center.name: "dc.01",
                ServicePropName.host_name.name: "zxcv-du",
                ServicePropName.ip_address.name: "ip.192.168.1.1",
                ServicePropName.service_name.name: "s_c",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "ccc",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.data_center.name: "dc.11",
                ServicePropName.host_name.name: "zxcv-dd",
                ServicePropName.ip_address.name: "ip.172.16.1.2",
                ServicePropName.service_name.name: "tt",
                ServicePropName.run_mode.name: "active",
                # FS_06_99_43_60 providing scalar value for list/array field is also possible:
                ServicePropName.group_label.name: "rrr",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.data_center.name: "dc.01",
                ServicePropName.host_name.name: "poiu-dd",
                ServicePropName.ip_address.name: "ip.192.168.1.3",
                ServicePropName.service_name.name: "xx",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "rrr",
                    "hhh",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-emea-upstream",
                ServicePropName.data_center.name: "dc.22",
                ServicePropName.host_name.name: "asdf-du",
                ServicePropName.ip_address.name: "ip.172.16.2.1",
                ServicePropName.service_name.name: "s_a",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "aaa",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-emea-upstream",
                ServicePropName.data_center.name: "dc.22",
                ServicePropName.host_name.name: "asdf-du",
                ServicePropName.ip_address.name: "ip.172.16.2.1",
                ServicePropName.service_name.name: "s_b",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "bbb",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.data_center.name: "dc.02",
                ServicePropName.host_name.name: "xcvb-dd",
                ServicePropName.ip_address.name: "ip.192.168.2.2",
                ServicePropName.service_name.name: "xx",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "rrr",
                    "hhh",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.data_center.name: "dc.02",
                ServicePropName.host_name.name: "xcvb-dd",
                ServicePropName.ip_address.name: "ip.192.168.2.2",
                ServicePropName.service_name.name: "zz",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "rrr",
                    "hhh",
                    "odd",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "dev-amer-upstream",
                ServicePropName.data_center.name: "dc.03",
                ServicePropName.host_name.name: "qwer-du",
                ServicePropName.ip_address.name: "ip.192.168.3.1",
                ServicePropName.service_name.name: "s_a",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "aaa",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-apac-upstream",
                ServicePropName.data_center.name: "dc.04",
                ServicePropName.host_name.name: "hjkl-qu",
                ServicePropName.ip_address.name: "ip.192.168.4.1",
                ServicePropName.service_name.name: "s_a",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "aaa",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-apac-upstream",
                ServicePropName.data_center.name: "dc.04",
                ServicePropName.host_name.name: "hjkl-qu",
                ServicePropName.ip_address.name: "ip.192.168.4.1",
                ServicePropName.service_name.name: "s_b",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "bbb",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-apac-upstream",
                ServicePropName.data_center.name: "dc.44",
                ServicePropName.host_name.name: "poiu-qu",
                ServicePropName.ip_address.name: "ip.172.16.4.2",
                ServicePropName.service_name.name: "s_c",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "ccc",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "qa-amer-upstream",
                ServicePropName.data_center.name: "dc.06",
                ServicePropName.host_name.name: "rtyu-qu",
                ServicePropName.ip_address.name: "ip.192.168.6.1",
                ServicePropName.service_name.name: "s_a",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "aaa",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-amer-downstream",
                ServicePropName.data_center.name: "dc.06",
                ServicePropName.host_name.name: "sdfgh-qd",
                ServicePropName.ip_address.name: "ip.192.168.6.3",
                ServicePropName.service_name.name: "tt1",
                ServicePropName.run_mode.name: "active",
                # FS_06_99_43_60 providing scalar value for list/array field is also possible:
                ServicePropName.group_label.name: "rrr",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-amer-downstream",
                ServicePropName.data_center.name: "dc.06",
                ServicePropName.host_name.name: "sdfgb-qd",
                ServicePropName.ip_address.name: "ip.192.168.6.4",
                ServicePropName.service_name.name: "xx",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "rrr",
                    "hhh",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "qwer-pd-1",
                ServicePropName.ip_address.name: "ip.192.168.7.1",
                ServicePropName.service_name.name: "s_a",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "aaa",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "qwer-pd-1",
                ServicePropName.ip_address.name: "ip.192.168.7.1",
                ServicePropName.service_name.name: "s_b",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "bbb",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "qwer-pd-3",
                ServicePropName.ip_address.name: "ip.192.168.7.2",
                ServicePropName.service_name.name: "s_c",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "ccc",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.77",
                ServicePropName.host_name.name: "qwer-pd-2",
                ServicePropName.ip_address.name: "ip.172.16.7.2",
                ServicePropName.service_name.name: "s_a",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "aaa",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.77",
                ServicePropName.host_name.name: "qwer-pd-2",
                ServicePropName.ip_address.name: "ip.172.16.7.2",
                ServicePropName.service_name.name: "s_b",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "bbb",
                    "xxx",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "upstream",
                ServicePropName.cluster_name.name: "prod-apac-upstream",
                ServicePropName.data_center.name: "dc.77",
                ServicePropName.host_name.name: "qwer-pd-2",
                ServicePropName.ip_address.name: "ip.172.16.7.2",
                ServicePropName.service_name.name: "s_c",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "ccc",
                    "sss",
                ],
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "prod-apac-downstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "wert-pd-1",
                ServicePropName.ip_address.name: "ip.192.168.7.3",
                ServicePropName.service_name.name: "tt1",
                ServicePropName.run_mode.name: "active",
                # FS_06_99_43_60 providing scalar value for list/array field is also possible:
                ServicePropName.group_label.name: "rrr",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "prod-apac-downstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "wert-pd-1",
                ServicePropName.ip_address.name: "ip.192.168.7.3",
                ServicePropName.service_name.name: "tt2",
                ServicePropName.run_mode.name: "passive",
                # FS_06_99_43_60 providing scalar value for list/array field is also possible:
                ServicePropName.group_label.name: "rrr",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "prod-apac-downstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "wert-pd-2",
                ServicePropName.ip_address.name: "ip.192.168.7.4",
                ServicePropName.service_name.name: "tt1",
                ServicePropName.run_mode.name: "passive",
                # FS_06_99_43_60 providing scalar value for list/array field is also possible:
                ServicePropName.group_label.name: "rrr",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "prod-apac-downstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "wert-pd-2",
                ServicePropName.ip_address.name: "ip.192.168.7.4",
                ServicePropName.service_name.name: "tt2",
                ServicePropName.run_mode.name: "active",
                # FS_06_99_43_60 providing scalar value for list/array field is also possible:
                ServicePropName.group_label.name: "rrr",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_63_37_05_36",  # demo
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                ServicePropName.code_maturity.name: "prod",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "prod-apac-downstream",
                ServicePropName.data_center.name: "dc.07",
                ServicePropName.host_name.name: "wert-pd-2",
                ServicePropName.ip_address.name: "ip.192.168.7.4",
                ServicePropName.service_name.name: "xx",
                ServicePropName.run_mode.name: "active",
                ServicePropName.group_label.name: [
                    "rrr",
                    "hhh",
                ],
            },
        ])

    def populate_TD_76_09_29_31_overlapped(
        self,
        cluster_envelopes: list[dict],
        host_envelopes: list[dict],
        service_envelopes: list[dict],
    ):
        if not self.is_test_data_allowed("TD_76_09_29_31"):
            return

        cluster_envelopes.extend([

            ############################################################################################################
            # TD_76_09_29_31 # overlapped: clusters

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-amer-downstream",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
            },
        ])

        host_envelopes.extend([

            ############################################################################################################
            # TD_76_09_29_31 # overlapped: hosts

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                # TD_76_09_29_31 geo_region set overlaps with host_name set:
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-amer-downstream",
                # TD_76_09_29_31 host_name set overlaps with geo_region set:
                ServicePropName.host_name.name: "amer",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-amer-downstream",
                # host_name is intentionally matching another host_name from different geo_region (also named alike):
                ServicePropName.host_name.name: "emea",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "amer",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-amer-downstream",
                # host_name is intentionally matching another host_name from different geo_region (also named alike):
                ServicePropName.host_name.name: "host-3-amer",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                # TD_76_09_29_31 geo_region set overlaps with host_name set:
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                # TD_76_09_29_31 host_name set overlaps with geo_region set:
                ServicePropName.host_name.name: "emea",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                # host_name is intentionally matching another host_name from different geo_region (also named alike):
                ServicePropName.host_name.name: "amer",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_76_09_29_31",  # overlapped
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                # host_name is intentionally matching another host_name from different geo_region (also named alike):
                ServicePropName.host_name.name: "host-3-emea",
            },
        ])

    def populate_TD_99_99_88_75_mutually_exclusive(
        self,
        cluster_envelopes: list[dict],
        host_envelopes: list[dict],
        service_envelopes: list[dict],
    ):
        if not self.is_test_data_allowed("TD_99_99_88_75"):
            return

        cluster_envelopes.extend([

            ############################################################################################################
            # TD_99_99_88_75 # mutually exclusive: clusters

            {
                envelope_payload_: {
                },
                test_data_: "TD_99_99_88_75",  # mutually exclusive
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-amer-downstream",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_99_99_88_75",  # mutually exclusive
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
            },
        ])

        host_envelopes.extend([

            ############################################################################################################
            # TD_99_99_88_75 # mutually exclusive: hosts

            {
                envelope_payload_: {
                },
                test_data_: "TD_99_99_88_75",  # mutually exclusive
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-apac-downstream",
                ServicePropName.host_name.name: "host-a-1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_99_99_88_75",  # mutually exclusive
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "qa",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "qa-apac-downstream",
                ServicePropName.host_name.name: "host-a-2",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_99_99_88_75",  # mutually exclusive
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.host_name.name: "host-b-1",
            },
            {
                envelope_payload_: {
                },
                test_data_: "TD_99_99_88_75",  # mutually exclusive
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.host_name.name: "host-b-2",
            },
        ])

    def populate_TD_38_03_48_51_large_generated(
        self,
        cluster_envelopes: list[dict],
        host_envelopes: list[dict],
        service_envelopes: list[dict],
    ):
        """
        TD_38_03_48_51: generate large data set
        """
        if not self.is_test_data_allowed("TD_38_03_48_51"):
            return

        for code_maturity in ["cm" + str(cmn) for cmn in range(0, self.object_multiplier)]:
            for geo_region in ["gr" + str(grn) for grn in range(0, self.object_multiplier)]:
                for flow_stage in ["fs" + str(fsn) for fsn in range(0, self.object_multiplier)]:

                    cluster_name = f"{code_maturity}-{geo_region}-{flow_stage}"

                    eprint(f"loading cluster_name={cluster_name}...")

                    #####################################################################################################
                    # clusters

                    generated_cluster = {
                        envelope_payload_: {
                        },
                        test_data_: "TD_38_03_48_51",  # large generated
                        ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                        ServicePropName.code_maturity.name: code_maturity,
                        ServicePropName.geo_region.name: geo_region,
                        ServicePropName.flow_stage.name: flow_stage,
                        ServicePropName.cluster_name.name: cluster_name,
                    }

                    cluster_envelopes.append(generated_cluster)

                    for host_name in ["hs" + str(hsn) for hsn in range(0, self.object_multiplier)]:

                        ################################################################################################
                        # hosts

                        generated_host = {
                            envelope_payload_: {
                            },
                            test_data_: "TD_38_03_48_51",  # large generated
                            ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                            ServicePropName.code_maturity.name: code_maturity,
                            ServicePropName.geo_region.name: geo_region,
                            ServicePropName.flow_stage.name: flow_stage,
                            ServicePropName.cluster_name.name: cluster_name,
                            ServicePropName.host_name.name: host_name,
                        }

                        host_envelopes.append(generated_host)

                        run_mode = "active"
                        for service_name in ["sn{:02d}".format(snn) for snn in range(0, self.object_multiplier)]:
                            ############################################################################################################
                            # services

                            generated_service = {
                                envelope_payload_: {
                                },
                                test_data_: "TD_38_03_48_51",  # large generated
                                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassService.name,
                                ServicePropName.code_maturity.name: code_maturity,
                                ServicePropName.geo_region.name: geo_region,
                                ServicePropName.flow_stage.name: flow_stage,
                                ServicePropName.cluster_name.name: cluster_name,
                                ServicePropName.host_name.name: host_name,
                                ServicePropName.service_name.name: service_name,
                                ServicePropName.run_mode.name: run_mode,
                            }

                            # Even (0, 2, ...) => active
                            # Odd (1, 3, ...) => passive
                            if run_mode == "active":
                                run_mode = "passive"
                            else:
                                run_mode = "active"

                            service_envelopes.append(generated_service)

    def populate_TD_43_24_76_58_single(
        self,
        cluster_envelopes: list[dict],
        host_envelopes: list[dict],
        service_envelopes: list[dict],
    ):
        if not self.is_test_data_allowed("TD_43_24_76_58"):
            return

        cluster_envelopes.extend([

            ############################################################################################################
            # TD_43_24_76_58 # single: clusters

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
            },
        ])

        host_envelopes.extend([

            ############################################################################################################
            # TD_43_24_76_58 # single: hosts

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.host_name.name: "qwer",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.host_name.name: "asdf",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.host_name.name: "qwer",
            },

            {
                envelope_payload_: {
                },
                test_data_: "TD_43_24_76_58",  # single
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.host_name.name: "asdf",
            },
        ])

    def populate_TD_39_25_11_76_missing_props(
        self,
        cluster_envelopes: list[dict],
        host_envelopes: list[dict],
        service_envelopes: list[dict],
    ):
        if not self.is_test_data_allowed("TD_39_25_11_76"):
            return

        cluster_envelopes.extend([

            ############################################################################################################
            # TD_39_25_11_76 # missing props: clusters

            {
                # nothing missing:
                envelope_payload_: {
                },
                test_data_: "TD_39_25_11_76",  # missing props
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
            },

            {
                # nothing missing:
                envelope_payload_: {
                },
                test_data_: "TD_39_25_11_76",  # missing props
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassCluster.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
            },

        ])

        host_envelopes.extend([

            ############################################################################################################
            # TD_39_25_11_76 # missing props: hosts

            {
                # nothing missing:
                envelope_payload_: {
                },
                test_data_: "TD_39_25_11_76",  # missing props
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.host_name.name: "qwer",
                ServicePropName.live_status.name: "green",
            },

            {
                # nothing missing:
                envelope_payload_: {
                },
                test_data_: "TD_39_25_11_76",  # missing props
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.host_name.name: "asdf",
                ServicePropName.live_status.name: "yellow",
            },

            {
                # missing `ServicePropName.live_status`:
                envelope_payload_: {
                },
                test_data_: "TD_39_25_11_76",  # missing props
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "apac",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-apac-downstream",
                ServicePropName.host_name.name: "zxcv",
                # NOTE: `ServicePropName.live_status` for this `data_envelope` is missing only in `apac` (not in `emea`).
                # ServicePropName.live_status.name: "red",
            },

            {
                # nothing missing:
                envelope_payload_: {
                },
                test_data_: "TD_39_25_11_76",  # missing props
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.host_name.name: "qwer",
                ServicePropName.live_status.name: "green",
            },

            {
                # nothing missing:
                envelope_payload_: {
                },
                test_data_: "TD_39_25_11_76",  # missing props
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.host_name.name: "asdf",
                ServicePropName.live_status.name: "yellow",
            },

            {
                # nothing missing:
                envelope_payload_: {
                },
                test_data_: "TD_39_25_11_76",  # missing props
                ReservedPropName.envelope_class.name: ServiceEnvelopeClass.ClassHost.name,
                ServicePropName.code_maturity.name: "dev",
                ServicePropName.geo_region.name: "emea",
                ServicePropName.flow_stage.name: "downstream",
                ServicePropName.cluster_name.name: "dev-emea-downstream",
                ServicePropName.host_name.name: "zxcv",
                ServicePropName.live_status.name: "red",
            },
        ])



from __future__ import annotations

from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.enum_desc.SpecialFunc import SpecialFunc
from argrelay.handler_response.DescribeLineArgsClientResponseHandler import DescribeLineArgsClientResponseHandler
from argrelay.plugin_delegator.AbstractJumpDelegator import AbstractJumpDelegator
from argrelay.runtime_context.InterpContext import function_container_ipos_
from argrelay.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_response.InvocationInput import InvocationInput

subsequent_function_container_ipos_ = function_container_ipos_ + 1


class QueryEnumDelegator(AbstractJumpDelegator):
    """
    FS_02_25_41_81: Implements `func_id_query_enum_items`.
    """

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        func_envelopes = [{
            instance_data_: {
                func_id_: SpecialFunc.func_id_query_enum_items.name,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedPropName.help_hint.name: "Enumerate available arg options (based on existing arg values)",
            ReservedPropName.func_state.name: FuncState.fs_alpha.name,
            ReservedPropName.func_id.name: SpecialFunc.func_id_query_enum_items.name,
        }]
        return func_envelopes

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        DescribeLineArgsClientResponseHandler.render_result(invocation_input)



class TestCase:
    """
    1.
    The input data provided by the `TestCase` (and its derived classes)
    parameterizes test scenario code and allows multiple test cases (e.g. via `TestCase.subTest`)
    run without changing that test scenario code.
    2.
    For test scenarios with variable input parameters, `TestCase` works as a builder pattern to provide:
    *   flexibility by omitting some parameters
        (assuming some defaults)
    *   flexibility by reordering parameters
        (e.g. instead of tuple with positional args, use named args or set them via func)
    *   flexibility by separating:
        *   partial input specification (e.g. common defaults)
        *   final input specification (e.g. individual overrides)
    3.
    Also, `TestCase` drives settings for any other `test_infra` classes, for example:
    *   Select specific `ServerActionVerifier` based on `CompType`.
    *   Init `EnvMockBuilder` mock the test input.
    *   Ensure input makes sense within given `*TestClass`.
    *   Ensure there is no missing or conflicting combination of input parameters.
    """

    def __init__(
        self,
        line_no: int,
        case_comment: str,
    ):
        self.line_no: int = line_no
        self.case_comment: str = case_comment

    def __str__(
        self,
    ):
        return f"line:{self.line_no}: {self.case_comment}"



from __future__ import annotations

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.enum_desc.SpecialChar import SpecialChar
from argrelay.enum_desc.TermColor import TermColor
from argrelay.handler_response.AbstractClientResponseHandler import AbstractClientResponseHandler
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay.schema_response.InterpResult import InterpResult
from argrelay.schema_response.InterpResultSchema import interp_result_desc

indent_size = 2


class DescribeLineArgsClientResponseHandler(AbstractClientResponseHandler):
    default_overrides_caption: str = "overrides"
    """
    FS_72_53_55_13: Caption for options hidden by defaults.
    """

    def __init__(
        self,
    ):
        super().__init__(
        )

    def handle_response(
        self,
        response_dict: dict,
    ):
        interp_result: InterpResult = interp_result_desc.dict_schema.load(response_dict)
        ElapsedTime.measure("after_object_creation")
        self.render_result(interp_result)

    @staticmethod
    def render_result(
        interp_result: InterpResult,
    ):
        """
        FS_02_25_41_81: Renders results `func_id_query_enum_items` (Alt+Shift+Q)
        """

        # TODO_11_77_28_50: print suggestions in `InterpResult.arg_values`

        print()

        # Print command line:
        for i in range(len(interp_result.all_tokens)):
            if i == interp_result.tan_token_ipos:
                DescribeLineArgsClientResponseHandler.render_tangent_token(interp_result, interp_result.all_tokens[i])
            elif i in interp_result.consumed_token_ipos_list():
                print(
                    f"{TermColor.consumed_token.value}{interp_result.all_tokens[i]}{TermColor.reset_style.value}",
                    end = " ",
                )
            elif i in interp_result.excluded_tokens:
                print(
                    f"{TermColor.excluded_token.value}{interp_result.all_tokens[i]}{TermColor.reset_style.value}",
                    end = " ",
                )
            else:
                print(
                    f"{TermColor.remaining_token.value}{interp_result.all_tokens[i]}{TermColor.reset_style.value}",
                    end = " ",
                )

        DescribeLineArgsClientResponseHandler.render_envelope_containers(
            interp_result.envelope_containers,
            interp_result.tan_token_l_part,
        )

    @staticmethod
    def render_tangent_token(
        interp_result: InterpResult,
        tan_token,
    ):
        """
        Implements FS_11_87_76_73 (highlight tangent prefix).
        """

        DescribeLineArgsClientResponseHandler.highlight_prefix(
            [tan_token],
            interp_result.tan_token_l_part,
            TermColor.tangent_token_r_part,
        )

    @staticmethod
    def render_envelope_containers(
        envelope_containers: list[EnvelopeContainer],
        value_prefix: str,
    ):
        print()
        is_first_missing_found: bool = False
        for envelope_container in envelope_containers:

            if envelope_container.found_count == 1:
                count_color = TermColor.found_count_1
            elif envelope_container.found_count > 1:
                count_color = TermColor.found_count_n
            else:
                count_color = TermColor.found_count_0
            print(
                f"{envelope_container.search_control.envelope_class}: {count_color.value}{envelope_container.found_count}{TermColor.reset_style.value}"
            )

            for key_to_type_dict in envelope_container.search_control.keys_to_types_list:
                arg_key = next(iter(key_to_type_dict))
                arg_type = key_to_type_dict[arg_key]

                if arg_type in envelope_container.assigned_types_to_values:
                    print(" " * indent_size, end = "")

                    # Set color based on `ArgSource`:
                    arg_source_color: TermColor
                    arg_source_color = DescribeLineArgsClientResponseHandler.select_arg_source_color(
                        envelope_container,
                        arg_type,
                    )
                    print(arg_source_color.value, end = "")

                    # Key = `arg_type`:
                    print(f"{arg_type}:", end = " ")

                    # Single value = `arg_value` (with prefix highlight):
                    DescribeLineArgsClientResponseHandler.highlight_prefix(
                        [envelope_container.assigned_types_to_values[arg_type].arg_value],
                        value_prefix,
                    )

                    # Restore color to name `ArgSource`:
                    print(arg_source_color.value, end = "")
                    print(
                        f"[{envelope_container.assigned_types_to_values[arg_type].arg_source.name}]",
                        end = ""
                    )
                    print(TermColor.reset_style.value, end = "")

                    # FS_72_53_55_13: Renders options hidden by default:
                    if len(envelope_container.filled_types_to_values_hidden_by_defaults) != 0:
                        if arg_type in envelope_container.filled_types_to_values_hidden_by_defaults:
                            print(" ", end = "")
                            print(TermColor.caption_hidden_by_default.value, end = "")
                            print(f"{DescribeLineArgsClientResponseHandler.default_overrides_caption}:", end = "")
                            print(TermColor.reset_style.value, end = " ")

                            values_hidden_by_defaults = envelope_container.filled_types_to_values_hidden_by_defaults[
                                arg_type
                            ]
                            DescribeLineArgsClientResponseHandler.highlight_prefix(
                                values_hidden_by_defaults,
                                value_prefix,
                                TermColor.value_hidden_by_default,
                            )

                elif arg_type in envelope_container.remaining_types_to_values:
                    print(" " * indent_size, end = "")

                    # Key = `arg_type`:
                    print(TermColor.remaining_value.value, end = "")
                    if not is_first_missing_found:
                        print(f"*{arg_type}:", end = "")
                        is_first_missing_found = True
                    else:
                        print(f"{arg_type}:", end = "")
                    print(f" ?", end = "")
                    print(TermColor.reset_style.value, end = " ")

                    # Multiple values = `arg_value` (with prefix highlight):
                    DescribeLineArgsClientResponseHandler.highlight_prefix(
                        envelope_container.remaining_types_to_values[arg_type],
                        value_prefix,
                    )

                else:
                    # The arg type is in the remaining but data have no arg values to suggest.
                    # Such arg types are shown because they are part of `search_control`.
                    # But they cannot be specified for current situation, otherwise, if already no data,
                    # any arg value assigned to such arg type would return no results.
                    print(" " * indent_size, end = "")
                    print(TermColor.no_option_to_suggest.value, end = "")
                    print(f"{arg_type}: ", end = "")
                    print(SpecialChar.NoPropValue.value, end = "")
                    print(TermColor.reset_style.value, end = "")

                print()

    @staticmethod
    def select_arg_source_color(
        envelope_container,
        arg_type,
    ):
        if envelope_container.assigned_types_to_values[arg_type].arg_source is ArgSource.ExplicitPosArg:
            arg_source_color = TermColor.explicit_pos_arg_value
        else:
            arg_source_color = TermColor.other_assigned_arg_value
        return arg_source_color

    @staticmethod
    def highlight_prefix(
        arg_values: list[str],
        value_prefix: str,
        default_color: TermColor = None,
    ):
        """
        Implements FS_32_05_46_00: highlight tangent prefix.
        """

        prefix_len: int = len(value_prefix)
        for arg_value in arg_values:
            # FS_32_05_46_00: use `startswith`
            if prefix_len > 0 and arg_value.startswith(value_prefix):
                print(
                    f"{TermColor.prefix_highlight.value}{TermColor.tangent_token_l_part.value}{arg_value[:prefix_len]}{TermColor.reset_style.value}",
                    # no space between L and R parts of the token:
                    end = "",
                )
                print(
                    f"{TermColor.tangent_token_r_part.value}{arg_value[prefix_len:]}{TermColor.reset_style.value}",
                    end = " ",
                )
            else:
                if default_color:
                    print(
                        f"{default_color.value}{arg_value}{TermColor.reset_style.value}",
                        end = " ",
                    )
                else:
                    print(
                        f"{arg_value}",
                        end = " ",
                    )



"""
This package generates OpenAPI spec definitions for `relay_server` which publishes them over HTTP.
While `relay_client` is supposed to use these definitions, it does not.
Instead, it uses `marshmallow` schemas these OpenAPI spec definitions are generated from.
"""

from __future__ import annotations

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin

from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.schema_request.CallContextSchema import call_context_desc
from argrelay.schema_response.ArgValuesSchema import arg_values_desc
from argrelay.schema_response.InterpResultSchema import interp_result_desc
from argrelay.schema_response.InvocationInputSchema import invocation_input_desc
from argrelay.server_spec.const_int import (
    DEFAULT_OPEN_API_VERSION,
    UNUSED_TITLE,
    UNUSED_VERSION,
)

# This spec is only used to generate data schemas: Marshmallow Schemas -> JSON Schemas.
# These data schemas are subsequently used in OAS specs for paths (which are defined manually).
# All serialization/deserialization/validation is done using `marshmallow` schemas (instead of JSON Schemas below).
server_op_data_schemas = APISpec(
    title = UNUSED_TITLE,
    version = UNUSED_VERSION,
    openapi_version = DEFAULT_OPEN_API_VERSION,
    plugins = [
        MarshmallowPlugin(
            # Override default resolver to preserve `Schema` postfix (keep full schema class name):
            schema_name_resolver = lambda schema_obj: schema_obj.__class__.__name__
        )
    ],
)


def add_type_desc_to_schema(type_desc: TypeDesc):
    """
    Generate data schemas: Marshmallow Schema (type_desc) -> JSON Schema
    """
    server_op_data_schemas.components.schema(
        type_desc.ref_name,
        schema = type_desc.dict_schema,
    )


def get_schema_definitions(ref_names: list[str]):
    """
    In order for `flasgger` to generate GUI, it needs full list schema definitions
    (including those added recursively for nested schemas) for Open API specs
    referenced via `@swag_from` annotations.
    At the moment, this list of schema names is composed manually
    (by checking errors on GUI or by running `test_auto_schema_is_in_response`).
    Actual schemas are generated via `marshmallow` + `apispec` inside:
    `server_op_data_schemas.components.schemas`
    This function (although providing minimum functionality) is the single point
    where this cumbersome mechanism is documented.
    """

    schema_definitions = {}
    for ref_name in ref_names:
        schema_definitions[ref_name] = server_op_data_schemas.components.schemas[ref_name]
    return schema_definitions


# Request: common for several API paths:
# *   ServerAction.DescribeLineArgs
# *   ServerAction.ProposeArgValues
# *   ServerAction.RelayLineArgs
add_type_desc_to_schema(call_context_desc)

# Response for:
# *   ServerAction.DescribeLineArgs:
add_type_desc_to_schema(interp_result_desc)

# Response for:
# *   ServerAction.ProposeArgValues:
add_type_desc_to_schema(arg_values_desc)

# Response for:
# *   ServerAction.RelayLineArgs:
add_type_desc_to_schema(invocation_input_desc)



from __future__ import annotations

from enum import Enum, auto

from argrelay.custom_integ.ServiceDelegator import set_default_to
from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.enum_desc.SpecialFunc import SpecialFunc
from argrelay.plugin_delegator.AbstractDelegator import get_func_id_from_invocation_input, get_func_id_from_interp_ctx
from argrelay.plugin_delegator.AbstractJumpDelegator import AbstractJumpDelegator
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_config_interp.SearchControlSchema import populate_search_control
from argrelay.schema_response.InvocationInput import InvocationInput
from argrelay.schema_response.InvocationInputSchema import invocation_input_desc

output_format_class_name = "OutputFormat"
output_format_prop_name = "output_format"

format_output_container_ipos_ = 1


class OutputFormat(Enum):
    """
    Output format values.
    """

    json_format = auto()
    repr_format = auto()
    text_format = auto()
    table_format = auto()


class InterceptDelegator(AbstractJumpDelegator):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        output_format_search_control = populate_search_control(
            class_to_collection_map,
            output_format_class_name,
            [
                {"output": output_format_prop_name},
            ],
        )

        func_envelopes = [{
            instance_data_: {
                func_id_: SpecialFunc.func_id_intercept_invocation.name,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    output_format_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedPropName.help_hint.name: (
                f"Intercept and print `{InvocationInput.__name__}` "
                "for specified function and its args"
            ),
            ReservedPropName.func_state.name: FuncState.fs_alpha.name,
            ReservedPropName.func_id.name: SpecialFunc.func_id_intercept_invocation.name,
        }]
        return func_envelopes

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def run_fill_control(
        self,
        interp_ctx: InterpContext,
    ) -> bool:
        func_id = get_func_id_from_interp_ctx(interp_ctx)
        any_assignment = False
        if func_id in [
            SpecialFunc.func_id_intercept_invocation.name,
        ]:
            # If we need to specify `output_format_class_name` `data_envelope`:
            if interp_ctx.curr_container_ipos == interp_ctx.curr_interp.base_container_ipos + format_output_container_ipos_:
                format_output_container = interp_ctx.envelope_containers[(
                    interp_ctx.curr_interp.base_container_ipos + format_output_container_ipos_
                )]
                any_assignment = (
                    set_default_to(output_format_prop_name, OutputFormat.json_format.name, format_output_container)
                    or
                    any_assignment
                )

        return any_assignment

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        # TODO: Fail (send to ErrorDelegator) if next function is not specified -
        #       showing the payload in this case is misleading.
        delegator_plugin_instance_id = self.plugin_instance_id
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.plugin_instance_entries[
                delegator_plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        # TODO: Print without first function `data_envelope` belonging to `intercept` function:
        func_id = get_func_id_from_invocation_input(invocation_input)
        if func_id == SpecialFunc.func_id_intercept_invocation.name:
            output_format: OutputFormat = OutputFormat[invocation_input.envelope_containers[
                format_output_container_ipos_
            ].data_envelopes[0][output_format_prop_name]]
            if not output_format:
                raise RuntimeError
            elif output_format == OutputFormat.json_format:
                print(invocation_input_desc.dict_schema.dumps(invocation_input))
            elif output_format == OutputFormat.repr_format:
                print(invocation_input)
            else:
                raise RuntimeError(f"not implemented: {output_format}")
        else:
            raise RuntimeError(f"unknown func_id: {func_id}")



from __future__ import annotations

from copy import deepcopy
from typing import Union

from argrelay.composite_tree.DictTreeWalker import surrogate_node_id_, fetch_subtree_node
from argrelay.enum_desc.InterpStep import InterpStep
from argrelay.misc_helper_server import insert_unique_to_sorted_list
from argrelay.plugin_interp.AbstractInterp import AbstractInterp
from argrelay.plugin_interp.InterpTreeInterpFactoryConfigSchema import interp_selector_tree_
from argrelay.plugin_interp.NoopInterpFactory import NoopInterpFactory
from argrelay.runtime_context.InterpContext import InterpContext


class InterpTreeInterp(AbstractInterp):
    """
    Implements FS_01_89_09_24 interp tree.
    """

    def __init__(
        self,
        interp_factory_id: str,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
    ):
        super().__init__(
            interp_factory_id,
            interp_tree_node_config_dict,
            interp_ctx,
        )
        self.interp_selector_tree: dict = interp_tree_node_config_dict[interp_selector_tree_]
        self.next_interp_factory_id: Union[str, None] = None

        # TODO: Why hard-coded? Isn't it possible for this plugin to be plugged into any depth of the tree?
        # Token with ipos = 0 is the command name eaten by `FirstArgInterp` (FS_42_76_93_51 first interp):
        self.base_token_ipos: int = 1

    def consumes_args_at_once(self) -> bool:
        return True

    def consume_pos_args(self) -> bool:
        """
        Consumes heading args in `remaining_arg_buckets` according to the `interp_selector_tree`.

        Unlike normal guideline to consume one arg at time (FS_44_36_84_88),
        this func consumes all possible args
        because args are selected according to the `interp_selector_tree` (not via query)
        and do not become incompatible by consuming all (causing FS_51_67_38_37 impossible arg combinations).

        Also, FS_01_89_09_24 interp tree does not observe FS_97_64_39_94 `arg_bucket` boundaries
        and consumes all necessary args sequentially.
        """

        curr_sub_tree = fetch_subtree_node(
            self.interp_selector_tree,
            self.interp_tree_abs_path,
        )
        any_consumed = False
        while True:
            if isinstance(curr_sub_tree, str):
                # Tree leaf is reached - use it:
                self.next_interp_factory_id = curr_sub_tree
                return any_consumed

            # Always consume next remaining token:
            # TODO: Is this assumption valid/safe that next remaining `ipos` is in the order it appears on command line?
            #       Apparently, it is fine as we keep deleting head of `remaining_arg_buckets` below:
            curr_token_ipos = self.interp_ctx.next_remaining_token_ipos()

            if curr_token_ipos is None:
                self.set_default_factory_id(curr_sub_tree)
                return any_consumed

            curr_token_value = self.interp_ctx.parsed_ctx.all_tokens[curr_token_ipos]
            assert self.is_pos_arg(curr_token_ipos)

            if isinstance(curr_sub_tree, dict):
                if curr_token_value in curr_sub_tree:
                    # Consume one more arg into path:
                    self.interp_tree_abs_path.append(curr_token_value)

                    bucket_index = self.interp_ctx.token_ipos_to_arg_bucket_map[curr_token_ipos]
                    self.interp_ctx.consumed_arg_buckets[bucket_index].append(curr_token_ipos)
                    del self.interp_ctx.remaining_arg_buckets[bucket_index][0]

                    curr_sub_tree = curr_sub_tree[curr_token_value]
                    any_consumed = True
                    continue
                else:
                    self.set_default_factory_id(curr_sub_tree)
                    return any_consumed
            else:
                raise LookupError()

    def set_default_factory_id(
        self,
        curr_sub_tree,
    ):
        # Impossible to consume more arg - use default of the current sub-tree:
        if surrogate_node_id_ in curr_sub_tree:
            self.next_interp_factory_id = curr_sub_tree[surrogate_node_id_]
        else:
            # TODO: TODO_62_75_33_41: Do not hardcode plugin instance id (instance of `NoopInterpFactory`):
            self.next_interp_factory_id = f"{NoopInterpFactory.__name__}.default"

    def try_iterate(self) -> InterpStep:
        return InterpStep.NextInterp

    def next_interp(self) -> "AbstractInterp":
        # Selected `next_interp_factory_id` need to be given one of the paths in the FS_01_89_09_24 interp tree
        # (because each `next_interp_factory_id` can be plugged into multiple interp tree leaves).
        self.interp_ctx.interp_tree_abs_path = tuple(self.interp_tree_abs_path)
        return self.interp_ctx.create_next_interp(self.next_interp_factory_id)

    def propose_arg_completion(self) -> None:
        """
        Suggest based on current sub-tree from config.
        """

        if not self.is_eligible_for_suggestion():
            return

        curr_sub_tree = fetch_subtree_node(
            self.interp_selector_tree,
            self.interp_tree_abs_path,
        )
        if isinstance(curr_sub_tree, dict):
            for proposed_value in [
                x for x in curr_sub_tree
                if (
                    isinstance(x, str)
                    and
                    x != surrogate_node_id_
                    and
                    # FS_32_05_46_00: using `startswith`:
                    x.startswith(self.interp_ctx.parsed_ctx.tan_token_l_part)
                )
            ]:
                insert_unique_to_sorted_list(self.interp_ctx.comp_suggestions, proposed_value)

    def is_eligible_for_suggestion(self):
        """
        Suggesting anything is possible only if there is no other tokens
        (after those recorded in `interp_tree_abs_path`) available for consumption by subsequent interpreters.
        """

        if self.interp_ctx.next_remaining_token_ipos() is None:
            return True
        else:
            return False

        # TODO: Clean up code below or take into account tangent token (ipos cursor position).
        #       Note that there is no ipos cursor position at the moment.
        #       It has to be computed with help of FS_23_62_89_43 tangent token
        #       (which might be a surrogate one if cursor does not touch non-whitespace chars).
        remaining_consumed_tokens = deepcopy(self.interp_ctx.consumed_token_ipos_list())

        # Remove everything until `base_token_ipos`:
        for token_ipos in range(0, self.base_token_ipos):
            assert token_ipos in remaining_consumed_tokens
            remaining_consumed_tokens.remove(token_ipos)

        # Remove everything eaten into `interp_tree_abs_path`:
        token_ipos = self.base_token_ipos
        for node_path_token in self.interp_tree_abs_path:
            assert token_ipos in remaining_consumed_tokens
            remaining_consumed_tokens.remove(token_ipos)
            token_ipos += 1

        # Is there anything else?
        return len(remaining_consumed_tokens) == 0



from datetime import datetime

from flask import Blueprint
from flask import render_template

from argrelay.relay_server.GuiBannerConfig import GuiBannerConfig
from argrelay.server_spec.const_int import (
    ARGRELAY_GUI_PATH,
    API_SPEC_PATH,
    API_DOCS_PATH,
)


def create_blueprint_gui(
    project_title: str,
    project_page_url: str,
    argrelay_version: str,
    gui_banner_config: GuiBannerConfig,
    server_start_time: int,
    project_git_commit_time: int,
    project_git_commit_url: str,
    project_git_commit_display_string: str,
    project_git_conf_dir_url: str,
    project_git_conf_dir_display_string: str,
):
    blueprint_gui = Blueprint(
        name = "blueprint_gui",
        import_name = __name__,
    )

    # TODO: test and think if it is the right place
    @blueprint_gui.route(ARGRELAY_GUI_PATH)
    def basic_ui():
        return render_template(
            "argrelay_main.html",
            project_title = project_title,
            project_page_url = project_page_url,
            argrelay_version = argrelay_version,
            argrelay_api_docs_path = API_DOCS_PATH,
            argrelay_api_spec_path = API_SPEC_PATH,
            server_start_time = unix_time_to_iso_utc(server_start_time),
            project_git_commit_time = unix_time_to_iso_utc(project_git_commit_time),
            project_git_commit_url = project_git_commit_url,
            project_git_commit_display_string = project_git_commit_display_string,
            project_git_conf_dir_url = project_git_conf_dir_url,
            project_git_conf_dir_display_string = project_git_conf_dir_display_string,
            header_html = gui_banner_config.header_html,
            footer_html = gui_banner_config.footer_html,
        )

    return blueprint_gui


def unix_time_to_iso_utc(
    unix_time: int,
) -> str:
    return f"{datetime.utcfromtimestamp(unix_time).isoformat()}Z"



from __future__ import annotations

from typing import Callable, Collection

from marshmallow import RAISE, fields

from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.EnvelopeCollection import EnvelopeCollection
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_interp.DataEnvelopeSchema import data_envelope_desc
from argrelay.schema_response.EnvelopeContainerSchema import data_envelopes_

index_props_ = "index_props"


class EnvelopeCollectionSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = EnvelopeCollection

    index_props = fields.List(
        fields.String(),
        required = False,
        load_default = [],
    )

    # TODO_00_79_72_55: do not store `data_envelopes`
    data_envelopes = fields.List(
        fields.Nested(data_envelope_desc.dict_schema),
        required = False,
        load_default = [],
    )


envelope_collection_desc = TypeDesc(
    dict_schema = EnvelopeCollectionSchema(),
    ref_name = EnvelopeCollectionSchema.__name__,
    dict_example = {
        index_props_: [
            "SomeTypeA",
            "SomeTypeB",
        ],
        data_envelopes_: [
            data_envelope_desc.dict_example,
        ],
    },
    default_file_path = "",
)


def init_envelop_collections(
    server_config: ServerConfig,
    class_names: Collection[str],
    get_index_props: Callable[[str, str], list[str]],
):
    """
    FS_56_43_05_79: Part of "search diff collection" implementation.

    Init:
    *   mapping from `class_name` to `collection_name`
    *   association of `collection_name` with its data as `EnvelopeCollection` (default = new and empty)
    *   list of `index_prop`-s for each `EnvelopeCollection` via `get_index_props`

    By default, class_name is mapped into collection_name which matches as string that class_name.
    """
    class_to_collection_map: dict = server_config.class_to_collection_map

    for class_name in class_names:
        # Default collection_name == class_name (unless overriden = specified in config explicitly):
        class_to_collection_map.setdefault(
            class_name,
            class_name,
        )
        collection_name = class_to_collection_map[class_name]
        envelope_collection = server_config.static_data.envelope_collections.setdefault(
            collection_name,
            EnvelopeCollection(
                index_props = [],
                data_envelopes = [],
            ),
        )

        index_props = envelope_collection.index_props
        for index_prop in get_index_props(collection_name, class_name):
            if index_prop not in index_props:
                index_props.append(index_prop)



from __future__ import annotations

from marshmallow import fields, RAISE, post_load

from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.runtime_data.PluginConfig import PluginConfig
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.schema_config_plugin.PluginEntrySchema import plugin_entry_desc

reusable_config_data_ = "reusable_config_data"
plugin_instance_entries_ = "plugin_instance_entries"


class PluginConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = PluginConfig

    # This `dict` provides a place to store arbitrary data.
    # YAML allows reusing any (substantially complex) data via aliases:
    # https://stackoverflow.com/a/48946813/441652
    # But `marshmallow` does not allow arbitrary data by default
    # (and it is kept that way to let garbage slip through).
    # This `dict` is ignored on load -
    # the data is used by YAML loader before it even reaches the schema validation.
    # See also values merge:
    # https://stackoverflow.com/a/46644785/441652
    reusable_config_data = fields.Dict(
        required = False,
        load_default = {},
    )

    # Plugin config data: key = `plugin_instance_id`, value = `plugin_entry`:
    plugin_instance_entries = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(plugin_entry_desc.dict_schema),
        required = True,
    )

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        # TODO: avoid calling it on load - it is only needed by server, not by client:
        # Build DAG:
        plugin_instance_id_activate_order_dag = {}
        for plugin_instance_id in input_dict[plugin_instance_entries_]:
            plugin_entry: PluginEntry = input_dict[plugin_instance_entries_][plugin_instance_id]
            plugin_instance_id_activate_order_dag[plugin_instance_id] = plugin_entry.plugin_dependencies

        return type(self).model_class(
            **input_dict,
            plugin_instance_id_activate_list = serialize_dag_to_list(
                plugin_instance_id_activate_order_dag,
            ),
        )


plugin_config_desc = TypeDesc(
    dict_schema = PluginConfigSchema(),
    ref_name = PluginConfigSchema.__name__,
    dict_example = {
        plugin_instance_entries_: {
            "some_plugin_instance_id": plugin_entry_desc.dict_example,
            "some_plugin_instance_id1": plugin_entry_desc.dict_example,
            "some_plugin_instance_id2": plugin_entry_desc.dict_example,
        },
    },
    default_file_path = "argrelay_plugin.yaml",
)


def serialize_dag_to_list(
    entire_dag: dict[str, list[str]],
) -> list[str]:
    output_list: list[str] = list_sub_dag(
        [],
        entire_dag,
        [node_id for node_id in entire_dag],
    )
    return output_list


def list_sub_dag(
    curr_path: list[str],
    entire_dag: dict[str, list[str]],
    id_sub_list: list[str],
) -> list[str]:
    output_list: list[str] = []

    for node_id in id_sub_list:
        if node_id in curr_path:
            raise ValueError(f"cyclic ref to plugin id in path `{curr_path}` -> `{node_id}`")
        if node_id not in entire_dag:
            raise ValueError(f"plugin id in path `{curr_path}` -> `{node_id}` is not defined")
        if node_id in entire_dag:
            # plugin has dependencies:
            sub_list = list_sub_dag(
                curr_path + [node_id],
                entire_dag,
                entire_dag[node_id],
            )
            for sub_node_id in sub_list:
                if sub_node_id not in output_list:
                    output_list.append(sub_node_id)
        else:
            raise ValueError(f"plugin id in path `{curr_path}` -> `{node_id}` is not included for activation")

        if node_id not in output_list:
            output_list.append(node_id)

    return output_list



from __future__ import annotations

import subprocess

from argrelay.custom_integ.BaseConfigDelegator import BaseConfigDelegator
from argrelay.custom_integ.ConfigOnlyDelegatorConfigSchema import (
    config_only_delegator_config_desc,
    command_template_,
    echo_command_on_stderr_,
)
from argrelay.misc_helper_common import eprint
from argrelay.plugin_delegator.AbstractDelegator import get_func_id_from_invocation_input
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_payload_,
)
from argrelay.schema_response.InvocationInput import InvocationInput

object_container_ipos_ = 1


class ConfigOnlyDelegator(BaseConfigDelegator):
    """
    Implements FS_49_96_50_77 config_only_delegator.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
            config_only_delegator_config_desc,
        )

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ):
        # TODO_74_73_60_93: Support expected envelope count in config-only delegator:
        # Use common static functions in `BaseConfigDelegator`.

        # Keep this as it is supposed to be referenced in the `command_template`
        envelope_containers = invocation_input.envelope_containers

        # There is no way to check on client side if this function belongs to this plugin
        # (client has no access to config):
        func_id = get_func_id_from_invocation_input(invocation_input)
        func_envelope = invocation_input.envelope_containers[0].data_envelopes[0]
        func_payload = func_envelope[envelope_payload_]
        command_template: str = func_payload[command_template_]

        # The input is trusted (from config), right? Then:
        # https://stackoverflow.com/a/54071505/441652
        command_line = eval(f'f"""\n{command_template}\n"""')

        if func_payload[echo_command_on_stderr_]:
            eprint(f"INFO: command_line:\n{command_line}")

        sub_proc = subprocess.run(
            command_line,
            shell = True,
        )
        exit(sub_proc.returncode)



import json
from dataclasses import asdict

import requests
from marshmallow import Schema

from argrelay.handler_response.AbstractClientResponseHandler import AbstractClientResponseHandler
from argrelay.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay.relay_client.AbstractClientCommand import AbstractClientCommand
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
from argrelay.schema_request.CallContextSchema import call_context_desc
from argrelay.server_spec.CallContext import CallContext
from argrelay.server_spec.const_int import BASE_URL_FORMAT


class AbstractRemoteClientCommand(AbstractClientCommand):

    def __init__(
        self,
        call_ctx: CallContext,
        connection_config: ConnectionConfig,
        response_handler: AbstractClientResponseHandler,
        response_schema: Schema,
        request_schema: Schema = call_context_desc.dict_schema,
    ):
        super().__init__(
            response_handler,
        )
        self.call_ctx: CallContext = call_ctx
        self.connection_config: ConnectionConfig = connection_config
        self.response_schema: Schema = response_schema
        self.request_schema: Schema = request_schema

    def execute_command(
        self,
    ):
        server_url = BASE_URL_FORMAT.format(**asdict(self.connection_config)) + f"{self.call_ctx.server_action.value}"
        headers_dict = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        request_json = self.request_schema.dumps(self.call_ctx)
        ElapsedTime.measure("before_request")
        response_obj = requests.post(
            server_url,
            headers = headers_dict,
            data = request_json,
        )
        ElapsedTime.measure("after_request")
        try:
            if response_obj.ok:
                # Leave both object creation and validation via schemas to `response_handler`.
                # Just deserialize into dict here:
                response_dict = json.loads(response_obj.text)
                ElapsedTime.measure("after_deserialization")
                self.response_handler.handle_response(response_dict)
            else:
                raise RuntimeError
        finally:
            ElapsedTime.measure("after_handle_response")



from enum import auto, Enum


class DistinctValuesQuery(Enum):
    """
    Selects implementation for FS_02_25_41_81 `func_id_query_enum_items`.

    See `test_QueryEngine_perf.py` for perf test used to produce these tables:

    Numbers for "use_single_collection: True" are more evident as `object_count` directly affects the query.

    Numbers for "use_single_collection: False" are less evident as `object_count` is distributed across collections.

    *   `mongomock`:

        ```
        ================================
        use_mongomock: True
        use_single_collection: True

                       object_multiplier         3         4         5         6         7         8         9
                            object_count       243      1024      3125      7776     16807     32768     59049
        --------------------------------
                  original_find_and_loop     0.012     0.044     0.134     0.459     1.715     6.159    28.553
                         native_distinct     0.077     0.209     0.581     1.538     3.186     6.295    11.231
                        native_aggregate     0.077     0.227     0.913     3.717    15.944    63.955   375.196
        ================================
        use_mongomock: True
        use_single_collection: False

                       object_multiplier         3         4         5         6         7         8         9
                            object_count       243      1024      3125      7776     16807     32768     59049
        --------------------------------
                  original_find_and_loop     0.011     0.031     0.102     0.431     1.750     6.061    35.473
                         native_distinct     0.057     0.172     0.468     1.247     2.938     5.556    10.528
                        native_aggregate     0.037     0.116     0.401     1.725     7.852    27.517   169.423
        ```

        *   The fastest on small collections is `original_find_and_loop`.
        *   The fastest on large collections is `native_distinct`.

    *   `pymongo`:

        ```
        ================================
        use_mongomock: False
        use_single_collection: True

                       object_multiplier         3         4         5         6         7         8         9
                            object_count       243      1024      3125      7776     16807     32768     59049
        --------------------------------
                  original_find_and_loop     0.028     0.034     0.070     0.147     0.255     0.493     1.040
                         native_distinct     0.070     0.137     0.264     0.550     0.976     1.896     3.452
                        native_aggregate     0.028     0.046     0.070     0.115     0.207     0.354     0.550
        ================================
        use_mongomock: False
        use_single_collection: False

                       object_multiplier         3         4         5         6         7         8         9
                            object_count       243      1024      3125      7776     16807     32768     59049
        --------------------------------
                  original_find_and_loop     0.033     0.027     0.055     0.141     0.197     0.342     0.744
                         native_distinct     0.047     0.088     0.158     0.285     0.479     0.821     1.389
                        native_aggregate     0.023     0.030     0.054     0.073     0.120     0.191     0.326
        ```

        The fastest is `native_aggregate`.
    """

    original_find_and_loop = auto()
    """
    See `QueryEngine._query_prop_values_original_find_and_loop`.

    It calls  MongoDB `find()` with `query_dict`, then loops to collect distinct values.

    This is the fastest option for small collections with `mongomock`.
    """

    native_distinct = auto()
    """
    See `QueryEngine._query_prop_values_native_distinct`.

    It does multiple calls to MongoDB `distinct()` with `query_dict`
    to collect set of distinct values per field at a time:
    https://stackoverflow.com/a/20655662/441652

    This is the fastest option for large collections with `mongomock`.
    """

    native_aggregate = auto()
    """
    See `QueryEngine._query_prop_values_native_aggregate`.

    It calls MongoDB `aggregate()` with `$match: query_dict`, then `$group` with `$addToSet`:
    *   See `test_MongoClient_distinc_values_search.py`:
        https://stackoverflow.com/a/63595331/441652
    *   Another example:
        https://stackoverflow.com/a/11991854/441652

    This is:
    *   the fastest option with `pymongo`.
    *   the slowest option with `mongomock`.
    """



from __future__ import annotations

from argrelay.composite_tree.CompositeNode import BaseNode, ZeroArgNode, InterpTreeNode, FuncTreeNode, TreePathNode


class AbstractNodeVisitor:

    def visit_node(
        self,
        node: BaseNode,
    ):
        if False:
            pass
        elif isinstance(node, ZeroArgNode):
            self._visit_ZeroArgNode(node)
        elif isinstance(node, TreePathNode):
            self._visit_TreePathNode(node)
        elif isinstance(node, InterpTreeNode):
            self._visit_InterpTreeNode(node)
        elif isinstance(node, FuncTreeNode):
            self._visit_FuncTreeNode(node)
        else:
            raise NotImplementedError(node.node_type.name)

    # noinspection PyPep8Naming
    def _visit_ZeroArgNode(self, node: ZeroArgNode) -> None:
        raise NotImplementedError()

    # noinspection PyPep8Naming
    def _visit_TreePathNode(self, node: TreePathNode) -> None:
        raise NotImplementedError()

    # noinspection PyPep8Naming
    def _visit_InterpTreeNode(self, node: InterpTreeNode) -> None:
        raise NotImplementedError()

    # noinspection PyPep8Naming
    def _visit_FuncTreeNode(self, node: FuncTreeNode) -> None:
        raise NotImplementedError()



from argrelay.client_command_remote.AbstractRemoteClientCommand import AbstractRemoteClientCommand
from argrelay.handler_response.ProposeArgValuesClientResponseHandler import ProposeArgValuesClientResponseHandler
from argrelay.runtime_data.ConnectionConfig import ConnectionConfig
from argrelay.schema_response.InterpResultSchema import interp_result_desc
from argrelay.server_spec.CallContext import CallContext


class ProposeArgValuesRemoteClientCommand(AbstractRemoteClientCommand):
    """
    This command is unused most of the time in favor of another (optimized):
    `ProposeArgValuesRemoteOptimizedClientCommand`

    See `completion_perf_notes.md` for details.

    This non-optimized implementation is still useful to get access to
    internal server state in tests via `LocalClient` - see FS_66_17_43_42 test_infra / special test mode #1.

    To enable use of this command, see:
    *   ClientConfig.optimize_completion_request
    *   LocalClientEnvMockBuilder
    """

    def __init__(
        self,
        call_ctx: CallContext,
        connection_config: ConnectionConfig,
    ):
        super().__init__(
            call_ctx = call_ctx,
            connection_config = connection_config,
            response_handler = ProposeArgValuesClientResponseHandler(),
            response_schema = interp_result_desc.dict_schema,
        )
